
REPOSITORY_QUEUE = "repositories"
PROJECT_QUEUE = "projects"

SECRET_SCANNER_TASK_NAME = "scan_repository"
LOGGING_FILE = "logging.ini"
LOG_FILE_PATH_PRJ_COLLECTOR = "project_collector.log"
LOG_FILE_PATH_REPO_COLLECTOR = "repository_collector.log"

BITBUCKET = "BITBUCKET"
AZURE_DEVOPS = "AZURE_DEVOPS"
GITHUB_PUBLIC = "GITHUB_PUBLIC"




# Standard Library
import json
import logging
from json import JSONDecodeError
from typing import List

# Third Party
from pydantic import ValidationError

# First Party
from vcs_scraper.model import VCSInstance

logger = logging.getLogger(__name__)


def parse_vcs_instances_file(filepath: str) -> List[VCSInstance]:
    vcs_instances: List[VCSInstance] = []
    errors_found = False
    logger.info(f"Reading VCS instances from file {filepath}")
    try:
        with open(filepath, encoding="utf-8") as vcs_instances_file:
            parsed_vcs_instances = json.loads(vcs_instances_file.read())
            logger.info(f"Parsing VCS instance definitions from file {filepath}")
            for vcs_instance in parsed_vcs_instances:
                try:
                    logger.info(f"Parsing VCS instance '{vcs_instance}'")
                    vcs_instances.append(VCSInstance(**parsed_vcs_instances[vcs_instance]))
                except ValidationError as validation_error:
                    logger.error(f"Failed while parsing VCS instance '{vcs_instance}': {validation_error}")
                    errors_found = True
    except JSONDecodeError as json_error:
        logger.error(f"Failed to parse VCS instances file '{filepath}': {json_error}")
        errors_found = True
    except FileNotFoundError:
        logger.error(f"VCS Instances file not found: '{filepath}'")
        errors_found = True
    if errors_found:
        return []
    return vcs_instances



# Standard Library
import logging

# Third Party
from tenacity import RetryError, Retrying, stop_after_attempt, wait_exponential

# First Party
from vcs_scraper.common import create_celery_client, initialise_logs, load_vcs_instances_into_map
from vcs_scraper.configuration import (
    DEBUG_MODE,
    RABBITMQ_DEFAULT_VHOST,
    RABBITMQ_QUEUES_PASSWORD,
    RABBITMQ_QUEUES_USERNAME,
    REQUIRED_ENV_VARS,
    RESC_RABBITMQ_SERVICE_HOST,
    VCS_INSTANCES_FILE_PATH
)
from vcs_scraper.constants import LOG_FILE_PATH_PRJ_COLLECTOR, PROJECT_QUEUE
from vcs_scraper.environment_wrapper import validate_environment
from vcs_scraper.vcs_connectors.vcs_connector_factory import VCSConnectorFactory
from vcs_scraper.vcs_instances_parser import VCSInstance

env_variables = validate_environment(REQUIRED_ENV_VARS)
logger_config = initialise_logs(LOG_FILE_PATH_PRJ_COLLECTOR, env_variables[DEBUG_MODE])
logger = logging.getLogger(__name__)


def collect_projects_from_vcs_instance(vcs_instance: VCSInstance) -> None:
    """
    Collects a list of all projects in the provided vcs instance and sends them to a RabbitMQ as celery tasks.
    """

    destination_message_queue = PROJECT_QUEUE

    task_name = "vcs_scraper.repository_collector.common.collect_repositories"
    vcs_client = VCSConnectorFactory.create_client_from_vcs_instance(vcs_instance)
    celery_client = create_celery_client("project_collector", env_variables[RABBITMQ_QUEUES_USERNAME],
                                         env_variables[RABBITMQ_QUEUES_PASSWORD],
                                         env_variables[RESC_RABBITMQ_SERVICE_HOST],
                                         env_variables[RABBITMQ_DEFAULT_VHOST])
    projects_to_scrape = []
    if vcs_instance.scope:
        logger.info(f"Checking if all projects within the scope exist in {vcs_client.url}")
        for project_in_scope in vcs_instance.scope:
            if vcs_client.project_exists(project_in_scope):
                logger.debug(f"Project '{project_in_scope}' is present in {vcs_client.url}")
                projects_to_scrape.append(project_in_scope)
            else:
                logger.warning(f"Project '{project_in_scope}' is not present in {vcs_client.url}")

    else:
        logger.info(f"Fetching list of all projects from {vcs_client.url}")
        projects_to_scrape = vcs_client.get_all_projects()
    projects_to_scrape_count = len(projects_to_scrape)
    logger.info(f"{projects_to_scrape_count} projects fetched successfully from '{vcs_client.url}'.")
    logger.info("Sending projects to the 'projects' queue.")
    projects_sent = 0
    for project_name in projects_to_scrape:
        if vcs_instance.exceptions:
            if project_name in vcs_instance.exceptions:
                logger.info(f"Skipping project {project_name} because it is within the exceptions.")
                continue
        try:
            for attempt in Retrying(wait=wait_exponential(multiplier=1, min=2, max=10), stop=stop_after_attempt(100)):
                with attempt:
                    try:
                        celery_client.send_task(task_name, kwargs={"project_key": project_name,
                                                                   "vcs_instance_name": vcs_instance.name},
                                                queue=destination_message_queue)
                    except Exception as exc:
                        logger.error(f"{exc} sending '{project_name}' to 'projects'. Retrying ...")
                        raise
            projects_sent += 1
            logger.info(f"Project '{project_name}' was sent successfully.")
        except RetryError as exc:
            raise SystemExit(f"Error while sending project '{project_name}' to the 'projects' : "
                             f"retry timed out") from exc

    logger.info(f"{projects_sent} Projects sent successfully to the 'projects' "
                f"queue out of {projects_to_scrape_count} projects.")


def collect_all_projects():

    vcs_instances_map = load_vcs_instances_into_map(env_variables[VCS_INSTANCES_FILE_PATH])

    for vcs_instance in vcs_instances_map:

        collect_projects_from_vcs_instance(vcs_instances_map[vcs_instance])



# coding=utf-8

# First Party
from vcs_scraper.environment_wrapper import EnvironmentVariable

RESC_RABBITMQ_SERVICE_HOST = "RESC_RABBITMQ_SERVICE_HOST"
RABBITMQ_DEFAULT_VHOST = "RABBITMQ_DEFAULT_VHOST"
RABBITMQ_QUEUES_USERNAME = "RABBITMQ_QUEUES_USERNAME"
RABBITMQ_QUEUES_PASSWORD = "RABBITMQ_QUEUES_PASSWORD"
PROJECT_QUEUE = "PROJECT_QUEUE"
VCS_INSTANCES_FILE_PATH = "VCS_INSTANCES_FILE_PATH"

DEBUG_MODE = "DEBUG_MODE"

REQUIRED_ENV_VARS = [

    EnvironmentVariable(
        DEBUG_MODE,
        "Show debug log statements, if set to '0' only INFO logs and above will be shown.",
        required=False,
        default="0",
    ),
    EnvironmentVariable(
        RABBITMQ_QUEUES_USERNAME,
        "The username used to connect to the rabbitmq project collector and repository collector topics.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_QUEUES_PASSWORD,
        "The password used to connect to the rabbitmq project collector and repository collector topics.",
        required=True,
    ),
    EnvironmentVariable(
        RESC_RABBITMQ_SERVICE_HOST,
        "The hostname/IP address of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_DEFAULT_VHOST,
        "The virtual host name of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        VCS_INSTANCES_FILE_PATH,
        "The absolute path to the json file containing the vcs_instances_definitions",
        required=True,
    ),
]



# Standard Library
import os
from typing import List

# Third Party
from github import Github
from github.Branch import Branch as GithubBranch
from github.Repository import Repository as GithubRepository

# First Party
from vcs_scraper.model import Branch, Repository, VCSInstance
from vcs_scraper.vcs_connectors.vcs_connector import VCSConnector


class GithubPublicConnector(VCSConnector):

    def project_exists(self, project_key: str) -> bool:
        return bool(self.api_client.get_user(project_key))

    def __init__(self, scheme, host, port, access_token, proxy=None):
        self.url = f"{scheme}://{host}:{port}"
        self.access_token = access_token
        self.proxy = proxy
        self._api_client = None

    @property
    def api_client(self):
        if not self._api_client:
            self._api_client = Github(login_or_token=self.access_token)
        return self._api_client

    def get_all_projects(self) -> List[str]:
        return [user.login for user in self.api_client.get_users()]

    def get_repos(self, project_key) -> List[dict]:
        repository_list = []
        repos = list(self.api_client.get_user(project_key).get_repos())
        for repo in repos:
            repo_details = self.get_repository_details(project_key=project_key, repository_name=repo.name)
            repo_dict = {
                "id": repo_details.id,
                "project_key": repo_details.full_name.split("/")[0],
                "name": repo_details.name,
                "html_url": repo_details.html_url
            }
            repository_list.append(repo_dict)
        return repository_list

    def get_repository_details(self, project_key: str, repository_name: str) -> GithubRepository:
        repo_details = self.api_client.get_repo(f"{project_key}/{repository_name}")
        return repo_details

    def get_branches(self, project_key: str, repository_id: str) -> List[GithubBranch]:
        branches = self.api_client.get_repo(f"{project_key}/{repository_id}").get_branches()
        for branch in branches:
            branch_details = self.get_branch_details(project_key=project_key, repository_name=repository_id,
                                                     branch=branch.name)
            branch.latest_commit = branch_details.commit.sha
        return branches

    def get_branch_details(self, project_key: str, repository_name: str, branch: str) -> GithubBranch:
        branch = self.api_client.get_repo(f"{project_key}/{repository_name}").get_branch(branch)
        return branch

    @staticmethod
    def export_repository(repository_information: GithubRepository, branches_information: List[GithubBranch],
                          vcs_instance_name: str) \
            -> Repository:
        """
        A method which generate a repository object about a single bitbucket repository.

        :param vcs_instance_name: Name of the VCS instance to which the repository belongs
        :param repository_information: Github repository information as returned by the Bitbucket API.
        :param branches_information: Github branches information for a single repo as returned by the Bitbucket API.
        :return Repository object
        """

        branches: List[Branch] = []
        for branch_information in branches_information:
            if os.getenv('SCAN_ONLY_MASTER_BRANCH', "true").lower() in "true":
                if branch_information.name.lower() in ["main", "master"]:
                    branch = Branch(repository_id=repository_information["id"],
                                    branch_name=branch_information.name,
                                    latest_commit=branch_information.latest_commit,
                                    branch_id=branch_information.name)
                    branches.append(branch)
                    break
            else:
                branch = Branch(repository_id=repository_information["id"],
                                branch_name=branch_information.name,
                                latest_commit=branch_information.latest_commit,
                                branch_id=branch_information.name)
                branches.append(branch)
        repository = Repository(branches=branches,
                                repository_url=repository_information["html_url"],
                                vcs_instance_name=vcs_instance_name,
                                repository_name=repository_information["name"],
                                repository_id=str(repository_information["id"]),
                                project_key=repository_information["project_key"])
        return repository

    @staticmethod
    def create_client_from_vcs_instance(vcs_instance: VCSInstance) -> VCSConnector:
        github_public_client = GithubPublicConnector(
            host=vcs_instance.hostname,
            scheme=vcs_instance.scheme,
            port=vcs_instance.port,
            access_token=vcs_instance.token
        )
        return github_public_client



# Standard Library
import logging.config
import sys
from distutils.sysconfig import get_python_lib
from os import path
from typing import Dict, List

# Third Party
from celery import Celery

# First Party
from vcs_scraper.model import VCSInstance
from vcs_scraper.vcs_instances_parser import parse_vcs_instances_file

logger = logging.getLogger(__name__)


def get_logging_settings_path():
    if path.isfile(get_python_lib() + "/vcs_scraper"):
        base_dir = get_python_lib() + "/vcs_scraper"
    else:
        base_dir = path.dirname(__file__)

    return base_dir + "/static/logging.ini"


def create_celery_client(queue_name: str, username: str, password: str, service_host: str, virtual_host: str):
    celery_client = Celery(queue_name,
                           broker=f"amqp://{username}:{password}@{service_host}/{virtual_host}")
    celery_client.conf.update({'worker_hijack_root_logger': False})
    celery_client.conf.update({'broker_connection_retry': True})
    celery_client.conf.update({'broker_connection_max_retries': 100})
    return celery_client


def initialise_logs(log_file_path: str, debug: str):
    logging_ini_file = get_logging_settings_path()
    logging.config.fileConfig(logging_ini_file, defaults={'log_file_path': log_file_path},
                              disable_existing_loggers=False)
    logger_config = logging.getLogger('root')
    if int(debug) == 1:
        logger_config.setLevel(logging.DEBUG)
    else:
        logger_config.setLevel(logging.INFO)
    return logger_config


def load_vcs_instances_into_map(file_path: str) -> Dict[str, VCSInstance]:
    vcs_instances: List[VCSInstance] = parse_vcs_instances_file(file_path)
    if not vcs_instances:
        logger.info("Exiting due to issues in VCS Instances definition in file "
                    f"{file_path}")
        sys.exit(-1)
    return {vcs_instance.name: vcs_instance for vcs_instance in vcs_instances}



# Standard Library
import abc
from typing import Dict, List

# First Party
from vcs_scraper.model import Repository, VCSInstance


class VCSConnector(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def api_client(self):
        pass

    @abc.abstractmethod
    def get_all_projects(self):
        pass

    @abc.abstractmethod
    def get_repos(self, project_key):
        pass

    @abc.abstractmethod
    def get_branches(self, project_key, repository_id):
        pass

    @staticmethod
    @abc.abstractmethod
    def export_repository(repository_information: Dict, branches_information: List[Dict],
                          vcs_instance_name: str) \
            -> Repository:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_client_from_vcs_instance(vcs_instance: VCSInstance):
        pass

    @abc.abstractmethod
    def project_exists(self, project_key: str) -> bool:
        pass



# First Party
from vcs_scraper.constants import AZURE_DEVOPS, BITBUCKET, GITHUB_PUBLIC
from vcs_scraper.vcs_connectors.azure_devops_connector import AzureDevopsConnector
from vcs_scraper.vcs_connectors.bitbucket_connector import BitbucketConnector
from vcs_scraper.vcs_connectors.github_public_connector import GithubPublicConnector
from vcs_scraper.vcs_instances_parser import VCSInstance


class VCSConnectorFactory:

    @staticmethod
    def create_client_from_vcs_instance(vcs_instance: VCSInstance):
        if vcs_instance.provider_type == BITBUCKET:
            return BitbucketConnector.create_client_from_vcs_instance(vcs_instance)
        if vcs_instance.provider_type == AZURE_DEVOPS:
            return AzureDevopsConnector.create_client_from_vcs_instance(vcs_instance)
        if vcs_instance.provider_type == GITHUB_PUBLIC:
            return GithubPublicConnector.create_client_from_vcs_instance(vcs_instance)
        raise NotImplementedError(f"Provider {vcs_instance.provider_type} is not supported")
